import java.awt.DisplayMode;
import java.awt.Font;
import java.awt.geom.Rectangle2D;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Insets;
import java.awt.Toolkit;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import javax.media.nativewindow.util.Dimension;
import javax.media.nativewindow.util.DimensionImmutable;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;

import javax.swing.JOptionPane;

import com.jogamp.newt.event.KeyEvent;
import com.jogamp.newt.event.KeyListener;
import com.jogamp.newt.MonitorDevice;
import com.jogamp.newt.MonitorMode;
import com.jogamp.newt.opengl.GLWindow;
import com.jogamp.newt.util.MonitorModeUtil;
import com.jogamp.opengl.util.awt.TextRenderer;
import com.jogamp.opengl.util.FPSAnimator;
import com.jogamp.opengl.util.texture.Texture;
import com.jogamp.opengl.util.texture.TextureData;

import javax.media.opengl.fixedfunc.GLMatrixFunc;
import javax.media.opengl.GL;
import javax.media.opengl.GL2;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLProfile;



class Tempest_in_a_Java_Pot_Window implements GLEventListener, KeyListener{

	private boolean should_Exit=false;
	private boolean settings_Were_Changed=false;


	// Settings that are specified by the user.
	private String audio_File_Path;
	private int AM_Frequency_To_Transmit_On;
	private int audio_Transmission_Amplification;
	private int algorithm_To_Use_For_Generating_Transmission_Frame;
	private int screen_Width, screen_Height;
	private float refresh_Rate;
	private boolean run_Full_Screen;

	// Settings that are derived from other things.
	private int window_Content_Width, window_Content_Height;
	private int pixel_Clock;
	private int pixels_Per_Horizontal_Blank;
	private int pixels_Per_AM_Cycle;
	private int sample_Rate;
	private int pixels_Per_Sample;
	private int samples_Per_Refresh;

	private Audio_Processing_Thread audio_Processing_Thread;
	private AudioInputStream audio_File_Audio_Input_Stream, downsampled_Audio_Input_Stream;
	private int downsampled_Audio_Input_Stream_Samples_Read;
	private SourceDataLine audio_Line;
	private GLWindow tempest_in_a_Java_Pot_Window;
	private TextRenderer settings_Text_Renderer;
	private int displayed_Frames_Count=0;

	private double[] sin_LUT=new double[256];
	private ByteBuffer texture_Buffer;
	private byte[] sine_Values, square_Values, half_Sine_Values, offset_Sine_Values, offset_Square_Values;
	private byte[] downsampled_Audio_Samples;

	private long time_Of_Last_User_Setting_Change=0;
	private boolean transmit_Tone=false;
	private int tone_Frequency=500;


	// If the constructor is called with a requested_Window_Content_Width and requested_Window_Content_Height then we
	// will run in a window.
	Tempest_in_a_Java_Pot_Window(String requested_Audio_File_Path, int requested_AM_Frequency_To_Transmit_On,
	  int requested_Audio_Transmission_Amplificaiton, int requested_Algorithm_To_Use_For_Generating_Transmission_Frame,
	  int requested_Screen_Width, int requested_Screen_Height, float requested_Refresh_Rate,
	  int requested_Window_Content_Width, int requested_Window_Content_Height){
		audio_File_Path=requested_Audio_File_Path;
		AM_Frequency_To_Transmit_On=requested_AM_Frequency_To_Transmit_On;
		audio_Transmission_Amplification=requested_Audio_Transmission_Amplificaiton;
		algorithm_To_Use_For_Generating_Transmission_Frame=requested_Algorithm_To_Use_For_Generating_Transmission_Frame;
		screen_Width=requested_Screen_Width;
		screen_Height=requested_Screen_Height;
		refresh_Rate=requested_Refresh_Rate;
		window_Content_Width=requested_Window_Content_Width;
		window_Content_Height=requested_Window_Content_Height;
		run_Full_Screen=false;
	}



	// If the constructor is called without a requested_Window_Content_Width and requested_Window_Content_Height then we
	// will run full screen.
	Tempest_in_a_Java_Pot_Window(String requested_Audio_File_Path, int requested_AM_Frequency_To_Transmit_On,
	  int requested_Audio_Transmission_Amplificaiton, int requested_Algorithm_To_Use_For_Generating_Transmission_Frame,
	  int requested_Screen_Width, int requested_Screen_Height, float requested_Refresh_Rate){
		audio_File_Path=requested_Audio_File_Path;
		AM_Frequency_To_Transmit_On=requested_AM_Frequency_To_Transmit_On;
		audio_Transmission_Amplification=requested_Audio_Transmission_Amplificaiton;
		algorithm_To_Use_For_Generating_Transmission_Frame=requested_Algorithm_To_Use_For_Generating_Transmission_Frame;
		screen_Width=requested_Screen_Width;
		screen_Height=requested_Screen_Height;
		refresh_Rate=requested_Refresh_Rate;
		window_Content_Width=requested_Screen_Width;
		window_Content_Height=requested_Screen_Height;
		run_Full_Screen=true;
	}



	public static List<DisplayMode> get_Available_Display_Modes(){
		List<DisplayMode> available_Display_Modes=new ArrayList<DisplayMode>();

		GraphicsDevice default_Screen_Device=GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
		DisplayMode[] all_Display_Modes=default_Screen_Device.getDisplayModes();

		// Sort all_Display_Modes from smallest pixel count to largest pixel count.
		Arrays.sort(all_Display_Modes, new Comparator<DisplayMode>() {
			@Override public int compare(DisplayMode first_Display_Mode, DisplayMode second_Display_Mode){
				return first_Display_Mode.getWidth()*first_Display_Mode.getHeight()-
				  second_Display_Mode.getWidth()*second_Display_Mode.getHeight();
			}
		});

		// Add each display_Mode that has a different resolution and refresh rate but not color depth which we don't
		// care about too much currently.
		for(int i=0; i<all_Display_Modes.length; i++){
			DisplayMode display_Mode=new DisplayMode(all_Display_Modes[i].getWidth(), all_Display_Modes[i].getHeight(),
			  0, all_Display_Modes[i].getRefreshRate());
			if(!available_Display_Modes.contains(display_Mode))
				available_Display_Modes.add(display_Mode);
		}

		return available_Display_Modes;
	}



	private void update_Derived_Variables_And_Data_Structures(){

		if(downsampled_Audio_Input_Stream!=null)
			sample_Rate=(int)downsampled_Audio_Input_Stream.getFormat().getSampleRate();


		int current_Screen_Width=0, current_Screen_Height=0, current_Refresh_Rate=0;
		if(tempest_in_a_Java_Pot_Window!=null){
			MonitorMode monitor_Mode=tempest_in_a_Java_Pot_Window.getMainMonitor().getCurrentMode();
			DimensionImmutable screen_Resolution=monitor_Mode.getSurfaceSize().getResolution();
			current_Screen_Width=screen_Resolution.getWidth();
			current_Screen_Height=screen_Resolution.getHeight();
			current_Refresh_Rate=(int)monitor_Mode.getRefreshRate();

			window_Content_Width=tempest_in_a_Java_Pot_Window.getWidth();
			window_Content_Height=tempest_in_a_Java_Pot_Window.getHeight();

			// Note: This may have been fixed in JOGL 2.2
			// tempest_in_a_Java_Pot_Window.getWidth() and tempest_in_a_Java_Pot_Window.getHeight() appear to return
			// wrong values on OS X when running full screen (it appears to return the width and height of the native
			// resolution of the screen, not the width and height of the current screen resolution). This appears to be
			// a bug in JOGL. Determine the correct values in an alternate fashion.
			if(tempest_in_a_Java_Pot_Window.isFullscreen()){
				window_Content_Width=current_Screen_Width;
				window_Content_Height=current_Screen_Height;
			}
		}


		if(window_Content_Width>=0 && window_Content_Height>=0)
			texture_Buffer=ByteBuffer.allocate(window_Content_Width*window_Content_Height);


		if(current_Screen_Width>0 && current_Screen_Height>0 && current_Refresh_Rate>0){
			double horizontal_Period_In_Microseconds=1000000.0/(current_Refresh_Rate*(1+current_Screen_Height+
			  Math.round((current_Screen_Height+1)/(1000000.0/(current_Refresh_Rate*550)-1.0))));
			double ideal_Duty_Cycle=30-3*horizontal_Period_In_Microseconds/10;

			pixels_Per_Horizontal_Blank=(int)(16*Math.round(current_Screen_Width*ideal_Duty_Cycle/
			  (1600-16*ideal_Duty_Cycle)));

			pixel_Clock=(int)((current_Screen_Width+pixels_Per_Horizontal_Blank)*1000000.0/
			  horizontal_Period_In_Microseconds);

			pixels_Per_AM_Cycle=pixel_Clock/AM_Frequency_To_Transmit_On+1;


			sine_Values=new byte[pixels_Per_AM_Cycle];
			for(int i=0; i<pixels_Per_AM_Cycle; i++)
				sine_Values[i]=(byte)(127*Math.sin(i*2*Math.PI/pixels_Per_AM_Cycle));

			square_Values=new byte[pixels_Per_AM_Cycle];
			for(int i=0; i<pixels_Per_AM_Cycle; i++)
				square_Values[i]=(byte)(i<pixels_Per_AM_Cycle/2 ? 127 : -127);

			half_Sine_Values=new byte[pixels_Per_AM_Cycle];
			for(int i=0; i<pixels_Per_AM_Cycle; i++)
				half_Sine_Values[i]=(byte)(Math.max(0, 255*Math.sin(i*2*Math.PI/pixels_Per_AM_Cycle)));

			offset_Sine_Values=new byte[pixels_Per_AM_Cycle];
			for(int i=0; i<pixels_Per_AM_Cycle; i++)
				offset_Sine_Values[i]=(byte)(128 + 127*Math.sin(i*2*Math.PI/pixels_Per_AM_Cycle));

			offset_Square_Values=new byte[pixels_Per_AM_Cycle];
			for(int i=0; i<pixels_Per_AM_Cycle; i++)
				offset_Square_Values[i]=(byte)(i<pixels_Per_AM_Cycle/2 ? 255 : 0);


			pixels_Per_Sample=pixel_Clock/sample_Rate+1;


			samples_Per_Refresh=(int)(sample_Rate/current_Refresh_Rate)+1;

			downsampled_Audio_Samples=new byte[samples_Per_Refresh];
		}
	}



	public void transmit_And_Play_Audio(){

		for(int i=0; i<256; i++)
			sin_LUT[i]=Math.sin(2*i*Math.PI/256);


		// Get the audio_File_Audio_Input_Stream.
		File audio_File=new File(audio_File_Path);
		try{
			audio_File_Audio_Input_Stream=AudioSystem.getAudioInputStream(audio_File);
		}catch(UnsupportedAudioFileException exception){
			JOptionPane.showMessageDialog(null, "The selected file is not an audio file type supported by this system.",
			  "Unsupported file type", JOptionPane.ERROR_MESSAGE);
			return;
		}catch(IOException exception){
			JOptionPane.showMessageDialog(null, "The selected file could not be read.", "Read error",
			  JOptionPane.ERROR_MESSAGE);
			return;
		}


		audio_Processing_Thread=new Audio_Processing_Thread(audio_File_Audio_Input_Stream);
		audio_Line=audio_Processing_Thread.audio_Line_To_Play_Audio_To;
		downsampled_Audio_Input_Stream=audio_Processing_Thread.audio_Input_Stream_To_Copy_Downsampled_Audio_To;
		downsampled_Audio_Input_Stream_Samples_Read=0;


		// Start the audio_Processing_Thread and wait until it is ready to proceed.
		synchronized(audio_Processing_Thread){
			audio_Processing_Thread.start();
			try{
				audio_Processing_Thread.wait();
			}catch(Exception e){
				System.out.println("An exception occurred while waiting for the audio processing thread to become " +
				  "ready.");
			}
		}


		tempest_in_a_Java_Pot_Window=GLWindow.create(new GLCapabilities(GLProfile.getDefault()));
		tempest_in_a_Java_Pot_Window.setTitle("Tempest in a Java Pot"); // TODO: Maybe put filename and/or AM frequency
		                                                                // in title too?
		tempest_in_a_Java_Pot_Window.addGLEventListener(this);
		tempest_in_a_Java_Pot_Window.addKeyListener(this);



		// TODO: Allow user to select monitor to run on?


		// It would be nice to make any requested display mode changes and have the window sized correctly before making
		// the window visible but unfortunately NEWT seems to require the window to be visible for the
		// setTopLevelPosition/Size() and getMainMonitor() functions to work correctly. Since the window needs to be
		// displayed it would be nice to make it a maximized window but NEWT also doesn't seem to allow you to maximize
		// a window. Alternatively you can calculate what size the window should be to fill up the screen (minus docks,
		// taskbars, etc...) and then set the window size to that. If it is important to have the window size and
		// position set correctly before displaying it, you probably could calculate the window insets using AWT and
		// then use setPosition/Size() taking the window insets into account. For now we'll just use a fixed 320x240
		// window intially and then resize it to the correct size in a bit.
		tempest_in_a_Java_Pot_Window.setSize(320, 240);
		tempest_in_a_Java_Pot_Window.setVisible(true);



		MonitorDevice monitor=tempest_in_a_Java_Pot_Window.getMainMonitor();

		// If the screen_Width and screen_Height are greater than 0 and refresh_Rate is greater than -1, then change to
		// the monitor_Mode specified by them.
		if(screen_Width>0 && screen_Height>0 && refresh_Rate>=0){
			// Get the available monitor_Modes for the monitor that tempest_in_a_Java_Pot_Window is on and filter the
			// list to those that match the screen_Width, screen_Height, and refresh_Rate.
			List<MonitorMode> monitor_Modes=monitor.getSupportedModes();
			monitor_Modes=MonitorModeUtil.filterByResolution(monitor_Modes, new Dimension(screen_Width, screen_Height));
			monitor_Modes=MonitorModeUtil.filterByRate(monitor_Modes, refresh_Rate);

			// Change the monitor mode to the first match.
			MonitorMode monitor_Mode=monitor_Modes.get(0);
			monitor.setCurrentMode(monitor_Mode);
		}


		// Update screen_Width, screen_Height, and refresh_Rate to match the current screen settings.
		DimensionImmutable screen_Resolution=monitor.getCurrentMode().getSurfaceSize().getResolution();
		screen_Width=screen_Resolution.getWidth();
		screen_Height=screen_Resolution.getHeight();
		refresh_Rate=monitor.getCurrentMode().getRefreshRate();


		// Go full screen if requested, otherwise just size the window to fill the available screen.
		if(run_Full_Screen){
			// I'm not sure but if the window is decorated when going full screen that may reduce the size of the
			// content area.
			tempest_in_a_Java_Pot_Window.setUndecorated(true);

			tempest_in_a_Java_Pot_Window.setFullscreen(true);
		}else{
			// Figure out the insets for the default screen device.
			Insets insets=Toolkit.getDefaultToolkit().getScreenInsets(
			  GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDefaultConfiguration());

			// Position and size the window so it uses up all available space on the screen minus any menubars,
			// taskbars, docks, etc...
			tempest_in_a_Java_Pot_Window.setTopLevelPosition(insets.left, insets.top);
			tempest_in_a_Java_Pot_Window.setTopLevelSize(screen_Width-insets.left-insets.right,
			  screen_Height-insets.top-insets.bottom);
		}


		tempest_in_a_Java_Pot_Window.requestFocus();


		update_Derived_Variables_And_Data_Structures();


		final FPSAnimator animator=new FPSAnimator(tempest_in_a_Java_Pot_Window, Math.round(refresh_Rate));
		animator.start();


		// Wait for audio_Processing_Thread to close.
		try{
			audio_Processing_Thread.join();
		}catch(Exception e){}


		if(animator.isAnimating()) animator.stop();
		tempest_in_a_Java_Pot_Window.destroy();
	}



	@Override public void keyPressed(KeyEvent e){
		switch(e.getKeyCode()){
			case KeyEvent.VK_ESCAPE:
			case KeyEvent.VK_Q:
				tempest_in_a_Java_Pot_Window.destroy();
				break;

			case KeyEvent.VK_LEFT:
			case KeyEvent.VK_A:
				AM_Frequency_To_Transmit_On-=1000;
				if(AM_Frequency_To_Transmit_On<1000) AM_Frequency_To_Transmit_On=1000;
				time_Of_Last_User_Setting_Change=System.currentTimeMillis();
				settings_Were_Changed=true;
				break;

			case KeyEvent.VK_RIGHT:
			case KeyEvent.VK_D:
				AM_Frequency_To_Transmit_On+=1000;
				if(AM_Frequency_To_Transmit_On>9999000) AM_Frequency_To_Transmit_On=9999000;
				time_Of_Last_User_Setting_Change=System.currentTimeMillis();
				settings_Were_Changed=true;
				break;

			case KeyEvent.VK_UP:
			case KeyEvent.VK_W:
				audio_Transmission_Amplification+=1;
				if(audio_Transmission_Amplification>400) audio_Transmission_Amplification=400;
				time_Of_Last_User_Setting_Change=System.currentTimeMillis();
				settings_Were_Changed=true;
				break;

			case KeyEvent.VK_DOWN:
			case KeyEvent.VK_S:
				audio_Transmission_Amplification-=1;
				if(audio_Transmission_Amplification<0) audio_Transmission_Amplification=0;
				time_Of_Last_User_Setting_Change=System.currentTimeMillis();
				settings_Were_Changed=true;
				break;

			case KeyEvent.VK_LESS:
			case KeyEvent.VK_COMMA:
				algorithm_To_Use_For_Generating_Transmission_Frame--;
				if(algorithm_To_Use_For_Generating_Transmission_Frame<0)
				  algorithm_To_Use_For_Generating_Transmission_Frame=0;
				time_Of_Last_User_Setting_Change=System.currentTimeMillis();
				settings_Were_Changed=true;
				break;

			case KeyEvent.VK_GREATER:
			case KeyEvent.VK_PERIOD:
				algorithm_To_Use_For_Generating_Transmission_Frame++;
				if(algorithm_To_Use_For_Generating_Transmission_Frame>6)
				  algorithm_To_Use_For_Generating_Transmission_Frame=6;
				time_Of_Last_User_Setting_Change=System.currentTimeMillis();
				settings_Were_Changed=true;
				break;

			case KeyEvent.VK_T:
				transmit_Tone=!transmit_Tone;
				break;

			case KeyEvent.VK_1:
			case KeyEvent.VK_NUMPAD1:
				tone_Frequency=100;
				break;

			case KeyEvent.VK_2:
			case KeyEvent.VK_NUMPAD2:
				tone_Frequency=200;
				break;

			case KeyEvent.VK_3:
			case KeyEvent.VK_NUMPAD3:
				tone_Frequency=300;
				break;

			case KeyEvent.VK_4:
			case KeyEvent.VK_NUMPAD4:
				tone_Frequency=400;
				break;

			case KeyEvent.VK_5:
			case KeyEvent.VK_NUMPAD5:
				tone_Frequency=500;
				break;

			case KeyEvent.VK_6:
			case KeyEvent.VK_NUMPAD6:
				tone_Frequency=600;
				break;

			case KeyEvent.VK_7:
			case KeyEvent.VK_NUMPAD7:
				tone_Frequency=700;
				break;

			case KeyEvent.VK_8:
			case KeyEvent.VK_NUMPAD8:
				tone_Frequency=800;
				break;

			case KeyEvent.VK_9:
			case KeyEvent.VK_NUMPAD9:
				tone_Frequency=900;
				break;

			case KeyEvent.VK_0:
			case KeyEvent.VK_NUMPAD0:
				tone_Frequency=1000;
				break;
		}
	}
	@Override public void keyReleased(KeyEvent e){};


	@Override public void init(GLAutoDrawable glDrawable){
		update_Derived_Variables_And_Data_Structures();

		GL2 gl=glDrawable.getGL().getGL2();

		gl.glMatrixMode(GLMatrixFunc.GL_PROJECTION);
		gl.glLoadIdentity();
		gl.glOrtho(0, (double)window_Content_Width, 0, (double)window_Content_Height, 0, 1);

		gl.glMatrixMode(GLMatrixFunc.GL_MODELVIEW);
		gl.glLoadIdentity();


		gl.glDisable(gl.GL_DEPTH_TEST);
		gl.setSwapInterval(1);

		settings_Text_Renderer=new TextRenderer(new Font("SansSerif", Font.PLAIN, 14), false, false);
	}


	@Override public void reshape(GLAutoDrawable glDrawable, int x, int y, int width, int height){
		update_Derived_Variables_And_Data_Structures();

		GL2 gl=glDrawable.getGL().getGL2();

		gl.glMatrixMode(GLMatrixFunc.GL_PROJECTION);
		gl.glLoadIdentity();
		gl.glOrtho(0, (double)window_Content_Width, 0, (double)window_Content_Height, 0, 1);

		gl.glMatrixMode(GLMatrixFunc.GL_MODELVIEW);
		gl.glLoadIdentity();
	}



	@Override public void display(GLAutoDrawable gLDrawable){

		// update_Derived_Variables_And_Data_Structures if settings_Were_Changed by the user.
		if(settings_Were_Changed){
			update_Derived_Variables_And_Data_Structures();
			settings_Were_Changed=false;
		}


		// Don't draw anything unless their is an animator and it is running.
		// TODO: Perhaps just redraw the last frame, just the AM carrier signal, or a blank frame.
		if(gLDrawable.getAnimator()==null || !gLDrawable.getAnimator().isAnimating())
			return;


		// Assuming this isn't the first call to display(), then at this point the frame that was processed on the
		// previous call to display() should now have just been put up on the display. The next frame that should be
		// displayed should normally be constructed using the audio data starting at the last position in
		// downsampled_Audio_Input_Stream. However if we're playing audio and the current audio playback time is more
		// than a certain amount off from the position in downsampled_Audio_Input_Stream (which could happen because
		// we're dropping frames, display() is being called more frequently than it is supposed to be, clock drift, or
		// just drift due to the sample rate not being evenly divisible by the number of samples in one frame duration)
		// in which case we'll use the audio data starting at the current audio playback time plus one frame duration.
		int audio_Sample_Index_To_Start_Processing_At=downsampled_Audio_Input_Stream_Samples_Read;
		if(audio_Line!=null && displayed_Frames_Count>0){
			int current_Audio_Line_Frame_Position=audio_Line.getFramePosition();

			// If the video is ahead of the audio then we won't read any additional data from
			// downsampled_Audio_Input_Stream and instead will redisplay the last frame. Later we should try to instead
			// still read some additional data from downsampled_Audio_Input_Stream but not as much in order to allow the
			// audio to catch up. The audio can be stretched a little by duplicating samples.
			if(current_Audio_Line_Frame_Position < (downsampled_Audio_Input_Stream_Samples_Read-samples_Per_Refresh) )
				audio_Sample_Index_To_Start_Processing_At=current_Audio_Line_Frame_Position+samples_Per_Refresh;

			// If the audio is ahead of the video then we'll just be able to skip ahead in the
			// downsampled_Audio_Input_Stream to the desired position.
			if( (downsampled_Audio_Input_Stream_Samples_Read+samples_Per_Refresh) < current_Audio_Line_Frame_Position)
				audio_Sample_Index_To_Start_Processing_At=current_Audio_Line_Frame_Position+samples_Per_Refresh;
		}


		// If we already displayed the first frame, then send a notification to the audio processing thread so it knows
		// to start audio playback.
		if(displayed_Frames_Count==1)
			audio_Processing_Thread.should_Play_Audio=true;


		// If necessary, advance the position in the downsampled_Audio_Input_Stream to
		// audio_Sample_Index_To_Start_Processing_At. Note that the samples are one byte each.
		if(audio_Sample_Index_To_Start_Processing_At>downsampled_Audio_Input_Stream_Samples_Read){
			try{
				downsampled_Audio_Input_Stream.skip(audio_Sample_Index_To_Start_Processing_At-
				  downsampled_Audio_Input_Stream_Samples_Read);
				downsampled_Audio_Input_Stream_Samples_Read=audio_Sample_Index_To_Start_Processing_At;
			}catch(Exception e){
				System.out.println("An exception occurred while syncing the downsampled audio stream position to the " +
				  "audio line.");
			}
		}


		// As long as the audio isn't behind the video, then get the next samples_Per_Refresh worth of downsampled audio
		// samples in order to use those audio samples later to generate the next frame to display.
		if(audio_Sample_Index_To_Start_Processing_At>=downsampled_Audio_Input_Stream_Samples_Read){
			try{
				for(int bytes_To_Copy=samples_Per_Refresh, total_Bytes_Read=0; bytes_To_Copy>0;){
					int bytes_Read=downsampled_Audio_Input_Stream.read(downsampled_Audio_Samples, total_Bytes_Read,
					  bytes_To_Copy);
					if(bytes_Read>=0){
						bytes_To_Copy-=bytes_Read;
						total_Bytes_Read+=bytes_Read;
						downsampled_Audio_Input_Stream_Samples_Read+=bytes_Read;
					}
				}


				// Replace the downsampled_Audio_Samples with a tone if we should transmit_Tone.
				if(transmit_Tone){
					for(int i=0; i<samples_Per_Refresh; i++)
						downsampled_Audio_Samples[i]=(byte)(128 + 127*Math.sin(i*2*Math.PI*tone_Frequency/sample_Rate));
				}


				// Amplify the downsampled_Audio_Samples by audio_Transmission_Amplification and clip the samples if
				// necessary.
				for(int i=0; i<samples_Per_Refresh; i++){
					int sample=((downsampled_Audio_Samples[i] & 0xff)-128) * audio_Transmission_Amplification / 100;
					if(sample<-128) sample=-128;
					if(sample>127) sample=127;
					sample+=128;
					downsampled_Audio_Samples[i]=(byte)sample;
				}
			}catch(Exception e){
				System.out.println("An exception occurred while reading the downsampled audio stream.");
			}
		}


		// Generate the data to use for the transmission frame texture.
		int full_Frame_Pixel_Index=0, frame_Sample_Index=0, pixel_Count_For_Current_Sample=0;
		for(int y=0; y<window_Content_Height; y++){
			for(int x=0; x<window_Content_Width; x++){

				// Theoretically method 0 should be the most correct for generating an AM signal. However for me even
				// method 6 which doesn't even bother trying to generate an AM carrier can be picked up on my radio when
				// set to the AM (but not FM) mode. I have a feeling that the AM circuitry is just picking up the audio
				// waveform but not so much the AM signal.
				// 0 = 128 + Sine wave AM carrier (-127 to 127) * Audio waveform (0 to 255) / 256
				// 1 = 128 + Sine wave AM carrier (-127 to 127) * Audio waveform (0 to 255) / 256
				//     (Faster but less precise method)
				// 2 = 128 + Square wave AM carrier (-127 to 127) * Audio waveform (0 to 255) / 256
				// 3 = Half sine wave AM carrier (0 to 255) * Audio waveform (0 to 255) / 256
				// 4 = Offset sine wave AM carrier (0 to 255) * Audio waveform (0 to 255) / 256
				// 5 = Offset square wave AM carrier (0 to 255) * Audio waveform (0 to 255) / 256
				// 6 = Audio waveform
				switch(algorithm_To_Use_For_Generating_Transmission_Frame){
					case 0:
						final double t=(x+y*(screen_Width+pixels_Per_Horizontal_Blank))/(double)pixel_Clock+
						  (displayed_Frames_Count%refresh_Rate)/(double)refresh_Rate;
						texture_Buffer.array()[y*window_Content_Width + x]=(byte)(128+
						  (downsampled_Audio_Samples[frame_Sample_Index] & 0xff)/2 *
						  sin_LUT[(int)(AM_Frequency_To_Transmit_On*256*t) % 256]
						  );
						break;

					case 1:
						texture_Buffer.array()[y*window_Content_Width + x]=(byte)(128+
						  (downsampled_Audio_Samples[frame_Sample_Index] & 0xff) *
						  sine_Values[full_Frame_Pixel_Index%pixels_Per_AM_Cycle] /
						  256);
						break;

					case 2:
						texture_Buffer.array()[y*window_Content_Width + x]=(byte)(128+
						  (downsampled_Audio_Samples[frame_Sample_Index] & 0xff) *
						  square_Values[full_Frame_Pixel_Index%pixels_Per_AM_Cycle] /
						  256);
						break;

					case 3:
						texture_Buffer.array()[y*window_Content_Width + x]=(byte)(
						  (downsampled_Audio_Samples[frame_Sample_Index] & 0xff) *
						  (half_Sine_Values[full_Frame_Pixel_Index%pixels_Per_AM_Cycle] & 0xff) /
						  256);
						break;

					case 4:
						texture_Buffer.array()[y*window_Content_Width + x]=(byte)(
						  (downsampled_Audio_Samples[frame_Sample_Index] & 0xff) *
						  (offset_Sine_Values[full_Frame_Pixel_Index%pixels_Per_AM_Cycle] & 0xff) /
						  256);
						break;

					case 5:
						texture_Buffer.array()[y*window_Content_Width + x]=(byte)(
						  (downsampled_Audio_Samples[frame_Sample_Index] & 0xff) *
						  (offset_Square_Values[full_Frame_Pixel_Index%pixels_Per_AM_Cycle] & 0xff) /
						  256);
						break;

					case 6:
						texture_Buffer.array()[y*window_Content_Width + x]=downsampled_Audio_Samples[frame_Sample_Index];
						break;
				}


				full_Frame_Pixel_Index++;
				pixel_Count_For_Current_Sample++;

				if(pixel_Count_For_Current_Sample==pixels_Per_Sample){
					frame_Sample_Index++;
					pixel_Count_For_Current_Sample=0;
				}
			}


			full_Frame_Pixel_Index+=pixels_Per_Horizontal_Blank+(screen_Width-window_Content_Width);
			pixel_Count_For_Current_Sample+=pixels_Per_Horizontal_Blank+(screen_Width-window_Content_Width);

			if(pixel_Count_For_Current_Sample>=pixels_Per_Sample){
				frame_Sample_Index+=pixel_Count_For_Current_Sample/pixels_Per_Sample;
				pixel_Count_For_Current_Sample=pixel_Count_For_Current_Sample%pixels_Per_Sample;
			}
		}


		//// Draw a white border just to assist with making sure the display is set up properly.
		//for(int x=0; x<window_Content_Width; x++){
		//	texture_Buffer.array()[0*window_Content_Width + x]=(byte)255;
		//	texture_Buffer.array()[(window_Content_Height-1)*window_Content_Width + x]=(byte)255;
		//}
		//for(int y=0; y<window_Content_Height; y++){
		//	texture_Buffer.array()[y*window_Content_Width + 0]=(byte)255;
		//	texture_Buffer.array()[y*window_Content_Width + window_Content_Width-1]=(byte)255;
		//}


		final GL2 gl = gLDrawable.getGL().getGL2();


		// TODO: I believe the texture code needs to be updated to work if NPOT textures aren't supported. See
		// http://jogamp.org/deployment/jogamp-next/javadoc/jogl/javadoc/com/jogamp/opengl/util/texture/Texture.html
		Texture texture=new Texture(gl, new TextureData(GLProfile.getDefault(), GL.GL_LUMINANCE, window_Content_Width,
		  window_Content_Height, 0, GL.GL_LUMINANCE, GL.GL_UNSIGNED_BYTE, false, false, false, texture_Buffer, null));
		texture.enable(gl);
		texture.bind(gl);


		gl.glBegin(GL2.GL_TRIANGLE_STRIP);
			gl.glTexCoord2f(0f, 0f); gl.glVertex3f(0f, (float)window_Content_Height, 0f);
			gl.glTexCoord2f(1f, 0f); gl.glVertex3f((float)window_Content_Width, (float)window_Content_Height, 0f);
			gl.glTexCoord2f(0f, 1f); gl.glVertex3f(0f, 0f, 0f);
			gl.glTexCoord2f(1f, 1f); gl.glVertex3f((float)window_Content_Width, 0f, 0f);
		gl.glEnd();


		texture.destroy(gl);


		// Display the current settings if the user changed them recently. The text and background could probably be
		// drawn more simply but this works for now.
		if(System.currentTimeMillis()-time_Of_Last_User_Setting_Change<2000){

			// Generate the text to display.
			String AM_Frequency_To_Transmit_On_String="AM frequency: " + AM_Frequency_To_Transmit_On/1000 + " kHz";
			String audio_Transmission_Amplification_String="Audio transmission amplification: " +
			  audio_Transmission_Amplification + "%";
			String algorithm_To_Use_For_Generating_Transmission_Frame_String_1="Algorithm to use for generating " +
			  "transmission frame:";
			String algorithm_To_Use_For_Generating_Transmission_Frame_String_2="";
			switch(algorithm_To_Use_For_Generating_Transmission_Frame){
				case 0:
					algorithm_To_Use_For_Generating_Transmission_Frame_String_2="Sine wave AM carrier x Audio " +
					  "waveform plus Offset";
					break;

				case 1:
					algorithm_To_Use_For_Generating_Transmission_Frame_String_2="Sine wave AM carrier x Audio " +
					  "waveform plus Offset (Faster but less precise)";
					break;

				case 2:
					algorithm_To_Use_For_Generating_Transmission_Frame_String_2="Square wave AM carrier x Audio " +
					  "waveform plus Offset";
					break;

				case 3:
					algorithm_To_Use_For_Generating_Transmission_Frame_String_2="Half sine wave AM carrier x Audio " +
					  "waveform";
					break;

				case 4:
					algorithm_To_Use_For_Generating_Transmission_Frame_String_2="Offset sine wave AM carrier x Audio " +
					  "waveform";
					break;

				case 5:
					algorithm_To_Use_For_Generating_Transmission_Frame_String_2="Offset square wave AM carrier x " +
					  "Audio waveform";
					break;

				case 6:
					algorithm_To_Use_For_Generating_Transmission_Frame_String_2="Audio waveform";
					break;
			}


			int settings_Text_Background_Box_Width=0, settings_Text_Background_Box_Height=0;
			final int BORDER_WIDTH=10, LINE_SPACING=5;
			Rectangle2D bounding_Box;


			// Calculate settings_Text_Background_Box_Width and settings_Text_Background_Box_Height.
			bounding_Box=settings_Text_Renderer.getBounds(AM_Frequency_To_Transmit_On_String);
			if(bounding_Box.getWidth()>settings_Text_Background_Box_Width)
				settings_Text_Background_Box_Width=(int)bounding_Box.getWidth();
			settings_Text_Background_Box_Height+=bounding_Box.getHeight();

			settings_Text_Background_Box_Height+=LINE_SPACING;

			bounding_Box=settings_Text_Renderer.getBounds(audio_Transmission_Amplification_String);
			if(bounding_Box.getWidth()>settings_Text_Background_Box_Width)
				settings_Text_Background_Box_Width=(int)bounding_Box.getWidth();
			settings_Text_Background_Box_Height+=bounding_Box.getHeight();

			settings_Text_Background_Box_Height+=LINE_SPACING;

			bounding_Box=settings_Text_Renderer.getBounds(algorithm_To_Use_For_Generating_Transmission_Frame_String_1);
			if(bounding_Box.getWidth()>settings_Text_Background_Box_Width)
				settings_Text_Background_Box_Width=(int)bounding_Box.getWidth();
			settings_Text_Background_Box_Height+=bounding_Box.getHeight();

			settings_Text_Background_Box_Height+=LINE_SPACING;

			bounding_Box=settings_Text_Renderer.getBounds(algorithm_To_Use_For_Generating_Transmission_Frame_String_2);
			if(bounding_Box.getWidth()>settings_Text_Background_Box_Width)
				settings_Text_Background_Box_Width=(int)bounding_Box.getWidth();
			settings_Text_Background_Box_Height+=bounding_Box.getHeight();

			settings_Text_Background_Box_Width+=2*BORDER_WIDTH;
			settings_Text_Background_Box_Height+=2*BORDER_WIDTH;


			// Draw a black settings_Text_Background_Box.
			gl.glBegin(GL2.GL_TRIANGLE_STRIP);
				gl.glColor3f(0f, 0f, 0f);
				gl.glVertex3f(window_Content_Width/2-settings_Text_Background_Box_Width/2,
				  window_Content_Height/2+settings_Text_Background_Box_Height/2, 0f);
				gl.glVertex3f(window_Content_Width/2+settings_Text_Background_Box_Width/2,
				  window_Content_Height/2+settings_Text_Background_Box_Height/2, 0f);
				gl.glVertex3f(window_Content_Width/2-settings_Text_Background_Box_Width/2,
				  window_Content_Height/2-settings_Text_Background_Box_Height/2, 0f);
				gl.glVertex3f(window_Content_Width/2+settings_Text_Background_Box_Width/2,
				  window_Content_Height/2-settings_Text_Background_Box_Height/2, 0f);
			gl.glEnd();


			// Draw the text from bottom to top.
			settings_Text_Renderer.beginRendering(window_Content_Width, window_Content_Height);
				int text_Vertical_Position=window_Content_Height/2-settings_Text_Background_Box_Height/2+BORDER_WIDTH;

				bounding_Box=settings_Text_Renderer.getBounds(
				  algorithm_To_Use_For_Generating_Transmission_Frame_String_2);
				settings_Text_Renderer.draw(algorithm_To_Use_For_Generating_Transmission_Frame_String_2,
				  window_Content_Width/2-(int)bounding_Box.getWidth()/2,
				  (int)(text_Vertical_Position+bounding_Box.getMaxY()));

				text_Vertical_Position+=(int)bounding_Box.getHeight()+LINE_SPACING;

				bounding_Box=settings_Text_Renderer.getBounds(
				  algorithm_To_Use_For_Generating_Transmission_Frame_String_1);
				settings_Text_Renderer.draw(algorithm_To_Use_For_Generating_Transmission_Frame_String_1,
				  window_Content_Width/2-(int)bounding_Box.getWidth()/2,
				  (int)(text_Vertical_Position+bounding_Box.getMaxY()));

				text_Vertical_Position+=(int)bounding_Box.getHeight()+LINE_SPACING;

				bounding_Box=settings_Text_Renderer.getBounds(audio_Transmission_Amplification_String);
				settings_Text_Renderer.draw(audio_Transmission_Amplification_String,
				  window_Content_Width/2-(int)bounding_Box.getWidth()/2,
				  (int)(text_Vertical_Position+bounding_Box.getMaxY()));

				text_Vertical_Position+=(int)bounding_Box.getHeight()+LINE_SPACING;

				bounding_Box=settings_Text_Renderer.getBounds(AM_Frequency_To_Transmit_On_String);
				settings_Text_Renderer.draw(AM_Frequency_To_Transmit_On_String,
				  window_Content_Width/2-(int)bounding_Box.getWidth()/2,
				  (int)(text_Vertical_Position+bounding_Box.getMaxY()));
			settings_Text_Renderer.endRendering();
		}


		displayed_Frames_Count++;
	}


	@Override public void dispose(GLAutoDrawable gLDrawable){
		should_Exit=true;
		audio_Processing_Thread.should_Play_Audio=false;
		audio_Processing_Thread.should_Exit=true;
		if(downsampled_Audio_Input_Stream!=null){
			try{
				downsampled_Audio_Input_Stream.close();
			}catch(Exception e){}
		}
	}
}